import { assert } from 'chai'
import {
  getPaginationParams,
  errorParser,
  removeEmptyStrings
} from '@/helpers'
import { createAxiosError } from 'axios-mock-adapter/src/utils'

describe('helpers', () => {
  describe('getPaginationParams', () => {
    it('returns parameters related to pagination in an URL', () => {
      const expected = {
        'https://example.page/?page=123': { page: '123' },
        'https://example.page/?with_count=true&cursor=ABC': { cursor: 'ABC' },
        'https://example.page/?with_count=true&page_size=100': {}
      }
      Object.entries(expected).forEach(([url, params]) => {
        assert.deepStrictEqual(getPaginationParams(url), params)
      })
    })
  })

  describe('errorParser', () => {
    it('should handle primitive types', () => {
      assert.strictEqual(errorParser(''), '')
      assert.strictEqual(errorParser(null), 'null')
      assert.strictEqual(errorParser(undefined), 'undefined')
      assert.strictEqual(errorParser(Infinity), 'Infinity')
      assert.strictEqual(errorParser(NaN), 'NaN')
      assert.strictEqual(errorParser(true), 'true')
      assert.strictEqual(errorParser(false), 'false')
    })
    it('should handle empty errors', () => {
      assert.strictEqual(errorParser(new Error()), 'Error')
    })
    it('should handle error messages', () => {
      assert.strictEqual(errorParser(new Error('Oh no!')), 'Oh no!')
    })
    it('should handle Axios errors with string responses', () => {
      const error = createAxiosError('Oh no!', null, { data: 'Some data' })
      assert.strictEqual(errorParser(error), 'Some data')
    })
    it('should handle Axios errors with array responses', () => {
      const error = createAxiosError('Oh no!', null, { data: ['Message 1', 'Message 2'] })
      assert.strictEqual(errorParser(error), 'Message 1 - Message 2')
    })
    it('should handle Axios errors with detail', () => {
      const error = createAxiosError('Oh no!', null, { data: { detail: ['Message 1', 'Message 2'] } })
      assert.strictEqual(errorParser(error), 'Message 1 - Message 2')
    })
    it('should handle Axios errors with unexpected responses', () => {
      const error = createAxiosError('Oh no!', null, { data: { thing1: 'Message 1', thing2: ['Message 2'] } })
      assert.strictEqual(errorParser(error), 'Oh no!')
    })
    it('should handle Axios errors with a general error', () => {
      const error = createAxiosError('Oh no!', null, { data: { __all__: ['Message 1', 'Message 2'] } })
      assert.strictEqual(errorParser(error), 'Message 1 - Message 2')
    })
  })

  describe('removeEmptyStrings', () => {
    it('should remove falsy values but not booleans', () => {
      assert.deepStrictEqual(removeEmptyStrings({}), {})
      assert.deepStrictEqual(removeEmptyStrings(
        {
          a: null,
          b: false,
          c: 4,
          d: undefined,
          e: NaN,
          f: 'something',
          g: ''
        }
      ), {
        b: false,
        c: 4,
        f: 'something'
      })
    })

    it('should handle non-object values', () => {
      assert.deepStrictEqual(removeEmptyStrings(), {})
      assert.deepStrictEqual(removeEmptyStrings(undefined), {})
      assert.deepStrictEqual(removeEmptyStrings(false), {})
      assert.deepStrictEqual(removeEmptyStrings(null), {})
    })
  })
})
