import axios from 'axios'
import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import { initialState, mutations } from '@/store/navigation'
import { useJobsStore } from '@/stores'
import store from './index.spec.js'
import { elementsSample, jobsSample } from '../samples'
import { FakeAxios, assertRejects } from '../testhelpers'

describe('navigation', () => {
  describe('mutations', () => {
    describe('setElements', () => {
      it('sets elements', () => {
        const state = { elements: null, scheduledDeletion: [] }
        mutations.setElements(state, { elements: elementsSample })
        assert.deepStrictEqual(state, { elements: elementsSample, scheduledDeletion: [] })
      })

      it('clears elements', () => {
        const state = { elements: elementsSample }
        mutations.setElements(state, { elements: null })
        assert.deepStrictEqual(state, { elements: null })
      })
    })

    it('addToScheduledDeletion', () => {
      const state = { elements: elementsSample, scheduledDeletion: [] }
      mutations.addToScheduledDeletion(state, 'volumeid')
      assert.deepStrictEqual(state, { elements: elementsSample, scheduledDeletion: ['volumeid'] })
    })

    describe('setIds', () => {
      it('sets the corpus and element IDs', () => {
        const state = initialState()
        mutations.setIds(state, { corpusId: 'corpusid', elementId: 'elementid' })
        assert.strictEqual(state.corpusId, 'corpusid')
        assert.strictEqual(state.elementId, 'elementid')
      })

      it('allows omitting the element ID', () => {
        const state = initialState()
        state.elementId = 'elementid'
        mutations.setIds(state, { corpusId: 'corpusid' })
        assert.strictEqual(state.corpusId, 'corpusid')
        assert.strictEqual(state.elementId, null)
      })
    })

    it('setFilters', () => {
      const state = initialState()
      mutations.setFilters(state, {
        folder: true,
        name: 'blah'
      })
      assert.deepStrictEqual(state.appliedFilters, {
        folder: true,
        name: 'blah'
      })
    })

    it('reset', () => {
      const state = { elements: elementsSample, scheduledDeletion: ['anid', 'anotherid'] }
      mutations.reset(state)
      assert.deepStrictEqual(state, initialState())
    })
  })

  describe('actions', () => {
    let mock, jobsStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      jobsStore = useJobsStore()
    })

    afterEach(() => {
      store.reset()
      jobsStore.$reset()
      // Remove any handlers, but leave mocking in place
      mock.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it('requires a corpus ID', async () => {
        assert.ok(!store.state.navigation.corpusId)
        const err = await assertRejects(async () => store.dispatch('navigation/list'))
        assert.strictEqual(err.toString(), 'Error: Missing corpus ID')
      })

      it('lists elements in a corpus', async () => {
        mock.onGet('/corpus/corpusid/elements/', {
          params: {
            top_level: true,
            with_corpus: false,
            with_classes: false,
            with_zone: false,
            page_size: 20
          }
        }).reply(200, elementsSample)

        store.state.corpora.corpora.corpusid = { id: 'corpusid' }
        store.state.navigation.corpusId = 'corpusid'
        await store.dispatch('navigation/list', { top_level: true })

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/list',
            payload: { top_level: true }
          },
          {
            mutation: 'navigation/setFilters',
            payload: { top_level: true }
          },
          {
            mutation: 'navigation/setElements',
            payload: { elements: null }
          },
          {
            action: 'corpora/get',
            payload: { id: 'corpusid' }
          },
          {
            mutation: 'navigation/setElements',
            payload: {
              elements: elementsSample,
              corpus: { id: 'corpusid' }
            }
          }
        ])

        assert.deepStrictEqual(store.state.navigation.appliedFilters, { top_level: true })
        assert.deepStrictEqual(store.state.navigation.elements, elementsSample)
      })

      it('lists elements in a parent', async () => {
        mock.onGet('/elements/elementid/children/', {
          params: {
            name: 'a',
            with_corpus: false,
            with_classes: false,
            with_zone: false,
            page_size: 20
          }
        }).reply(200, elementsSample)

        store.state.corpora.corpora.corpusid = { id: 'corpusid' }
        store.state.navigation.corpusId = 'corpusid'
        store.state.navigation.elementId = 'elementid'
        await store.dispatch('navigation/list', { name: 'a' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/list',
            payload: { name: 'a' }
          },
          {
            mutation: 'navigation/setFilters',
            payload: { name: 'a' }
          },
          {
            mutation: 'navigation/setElements',
            payload: { elements: null }
          },
          {
            action: 'corpora/get',
            payload: { id: 'corpusid' }
          },
          {
            mutation: 'navigation/setElements',
            payload: {
              elements: elementsSample,
              corpus: { id: 'corpusid' }
            }
          }
        ])

        assert.deepStrictEqual(store.state.navigation.appliedFilters, { name: 'a' })
        assert.deepStrictEqual(store.state.navigation.elements, elementsSample)
      })

      it('loads corpora when needed', async () => {
        const corpus = {
          id: 'corpusid',
          name: 'Le Corpus',
          rights: ['read'],
          types: []
        }
        mock.onGet('/corpus/').reply(200, [corpus])

        mock.onGet('/corpus/corpusid/elements/', {
          params: {
            top_level: true,
            with_corpus: false,
            with_classes: false,
            with_zone: false,
            page_size: 20
          }
        }).reply(200, elementsSample)

        store.state.navigation.corpusId = 'corpusid'
        await store.dispatch('navigation/list', { top_level: true })

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/list',
            payload: { top_level: true }
          },
          {
            mutation: 'navigation/setFilters',
            payload: { top_level: true }
          },
          {
            mutation: 'navigation/setElements',
            payload: { elements: null }
          },
          {
            action: 'corpora/get',
            payload: { id: 'corpusid' }
          },
          {
            action: 'corpora/$_fetch'
          },
          {
            mutation: 'corpora/set',
            payload: [corpus]
          },
          {
            mutation: 'navigation/setElements',
            payload: {
              elements: elementsSample,
              corpus: { ...corpus, types: {} }
            }
          }
        ])

        assert.deepStrictEqual(store.state.navigation.appliedFilters, { top_level: true })
        assert.deepStrictEqual(store.state.navigation.elements, elementsSample)
      })
    })

    describe('delete', () => {
      it('requires a corpus ID', async () => {
        assert.ok(!store.state.navigation.corpusId)
        const err = await assertRejects(async () => store.dispatch('navigation/delete'))
        assert.strictEqual(err.toString(), 'Error: Missing corpus ID')
      })

      it('deletes elements in a corpus', async () => {
        mock.onDelete('/corpus/corpusid/elements/', {
          params: {
            name: 'blah'
          }
        }).reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)

        store.state.corpora.corpora.corpusid = { id: 'corpusid' }
        store.state.navigation.corpusId = 'corpusid'
        store.state.navigation.appliedFilters = { name: 'blah' }
        await store.dispatch('navigation/delete')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/delete'
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'success',
              text: 'Element deletion has been scheduled.'
            }
          }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('delete elements in a parent', async () => {
        mock.onDelete('/elements/elementid/children/', {
          params: {
            name: 'blah'
          }
        }).reply(204)
        mock.onGet('/jobs/').reply(200, jobsSample)

        store.state.corpora.corpora.corpusid = { id: 'corpusid' }
        store.state.navigation.corpusId = 'corpusid'
        store.state.navigation.elementId = 'elementid'
        store.state.navigation.appliedFilters = { name: 'blah' }
        await store.dispatch('navigation/delete')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/delete'
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'success',
              text: 'Element deletion has been scheduled.'
            }
          }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('handles errors', async () => {
        mock.onDelete('/corpus/corpusid/elements/', {
          params: {
            name: 'blah'
          }
        }).reply(400, { detail: 'you must construct additional pylons' })
        mock.onGet('/jobs/').reply(200, jobsSample)

        store.state.corpora.corpora.corpusid = { id: 'corpusid' }
        store.state.navigation.corpusId = 'corpusid'
        store.state.navigation.appliedFilters = { name: 'blah' }
        await store.dispatch('navigation/delete')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'navigation/delete'
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'you must construct additional pylons'
            }
          }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })
    })
  })
})
