import { assert } from 'chai'
import axios from 'axios'
import { pick } from 'lodash'
import { useWorkerStore, useNotificationStore } from '@/stores'
import { FakeAxios, setUpTestPinia, actionsCompletedPlugin, assertRejects } from '../testhelpers'
import { DuplicatedWorker } from '@/api'
import { workerConfigurationsSample, workersSample, workerVersionsSample, workerSample, workerRunsSample } from '../samples'

describe('workers', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setUpTestPinia([actionsCompletedPlugin])
      store = useWorkerStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listWorkers', () => {
      it('lists workers for a repository', async () => {
        const worker1 = {
          id: 'workerid1',
          name: 'Worker 1',
          slug: 'worker_1',
          type: 'classifier'
        }
        const worker2 = {
          id: 'workerid2',
          name: 'Worker 2',
          slug: 'worker_2',
          type: 'classifier'
        }
        const reply = {
          count: 2,
          results: [worker1, worker2]
        }

        mock.onGet('/workers/', { page: 1 }).reply(200, reply)

        await store.listWorkers('repoid')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/'
          }
        ])
        assert.deepStrictEqual(store.workers, { workerid1: worker1, workerid2: worker2 })
      })
    })

    describe('listWorkerTypes', () => {
      it('lists available worker types on the instance', async () => {
        const workerType1 = {
          id: 'worker_type_1',
          display_name: 'Classifier',
          slug: 'classifier'
        }
        const workerType2 = {
          id: 'worker_type_2',
          display_name: 'OCR',
          slug: 'ocr'
        }
        const workerType3 = {
          id: 'worker_type_3',
          display_name: 'Document Layout Analysis',
          slug: 'dla'
        }

        mock.onGet('/workers/types/', { params: { page: 1 } }).reply(200, {
          count: 3,
          number: 1,
          next: 'http://something',
          results: [workerType1, workerType2]
        })
        mock.onGet('/workers/types/', { params: { page: 2 } }).reply(200, {
          count: 3,
          number: 2,
          next: null,
          results: [workerType3]
        })

        await store.listWorkerTypes()
        await store.actionsCompleted()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/types/'
          },
          {
            method: 'get',
            url: '/workers/types/'
          }
        ])
        assert.deepStrictEqual(store.workerTypes, { worker_type_1: workerType1, worker_type_2: workerType2, worker_type_3: workerType3 })
      })
    })

    describe('listConfigurations', () => {
      it('lists worker configurations for a specific worker', async () => {
        mock.onGet('/workers/workerid/configurations/', { params: { page: 1 } }).reply(200, { number: 1, next: 'http://url', results: [workerConfigurationsSample[0]] })
        mock.onGet('/workers/workerid/configurations/', { params: { page: 2 } }).reply(200, { number: 2, next: null, results: [workerConfigurationsSample[1]] })

        await store.listConfigurations('workerid', { page: 1 })
        await store.actionsCompleted()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/workerid/configurations/'
          },
          {
            method: 'get',
            url: '/workers/workerid/configurations/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, {
          workerid: {
            configid1: workerConfigurationsSample[0],
            configid2: workerConfigurationsSample[1]
          }
        })
      })

      it('handles HTTP errors', async () => {
        mock.onGet('/workers/workerid/configurations/').reply(500, { detail: ['Something went wrong'] })

        await store.listConfigurations('workerid', { page: 1 })

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Something went wrong'
          }
        ])
      })

      it('handles pagination error', async () => {
        mock.onGet('/workers/workerid/configurations/').reply(200, { results: workerConfigurationsSample })

        await store.listConfigurations('workerid', { page: 1 })
      })
    })

    describe('createConfiguration', () => {
      it('creates a worker configuration for a specific worker', async () => {
        const configuration = { ...workerConfigurationsSample[0] }
        delete configuration.id
        mock.onPost('/workers/workerid/configurations/').reply(200, { ...workerConfigurationsSample[0] })

        await store.createConfiguration('workerid', configuration)

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/workers/workerid/configurations/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, {
          workerid: {
            [workerConfigurationsSample[0].id]: workerConfigurationsSample[0]
          }
        })
      })
    })

    describe('getConfiguration', () => {
      it('retrieves a given worker configuration', async () => {
        const reply = {
          id: 'configid1',
          name: 'configname1',
          configuration: {
            test: 'test'
          }
        }

        mock.onGet('/workers/configurations/configid1/').reply(200, reply)

        await store.getConfiguration('workerid1', 'configid1')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/configurations/configid1/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, { workerid1: { configid1: reply } })
      })
    })

    describe('updateConfiguration', () => {
      it('updates a worker configuration', async () => {
        const response = {
          id: 'configid1',
          name: 'configname1',
          configuration: {
            test: 'test'
          },
          archived: true
        }

        /**
         * This must be set as if it is totally empty this.workerConfigurations[workerid1] is undefined
         * this.workerConfigurations[workerid1][configid1] cannot be set by updateConfiguration
         */
        store.workerConfigurations = {
          workerid1: {}
        }
        mock.onPatch('/workers/configurations/configid1/').reply(200, response)

        await store.updateConfiguration(
          'workerid1',
          'configid1',
          { name: 'blob' }
        )

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'patch',
            url: '/workers/configurations/configid1/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, {
          workerid1: {
            configid1: response
          }
        })
      })

      it('handles errors', async () => {
        mock.onPatch('/workers/configurations/configid1/').reply(500, { detail: ['An unexpected Spanish inquisition occurred.'] })

        await store.updateConfiguration(
          'workerid1',
          'configid1',
          { name: 'blob' }
        )

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'An unexpected Spanish inquisition occurred.'
          }
        ])
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'patch',
            url: '/workers/configurations/configid1/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, {})
      })

      it('removes a worker configuration when it is archived', async () => {
        const config = {
          id: 'configid1',
          name: 'configname1',
          configuration: {
            test: 'test'
          },
          archived: false
        }
        store.workerConfigurations.workerid1 = {
          configid1: config
        }

        mock.onPatch('/workers/configurations/configid1/').reply(200, {
          ...config,
          archived: true
        })

        await store.updateConfiguration(
          'workerid1',
          'configid1',
          { archived: true }
        )

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'patch',
            url: '/workers/configurations/configid1/'
          }
        ])
        assert.deepStrictEqual(store.workerConfigurations, {
          workerid1: {}
        })
      })
    })

    describe('listVersions', () => {
      it('lists versions for a specific worker', async () => {
        const version1 = {
          id: 'versionid1',
          configuration: {
            test: 'test'
          },
          revision: {
            id: 'revisionid1',
            state: 'created',
            hash: 'xxxxxxxxxxxxxxxx',
            author: 'Bob',
            message: 'commit message',
            created: '2020-05-27',
            commit_url: 'https://xxxxxxxxxxxxxxxx',
            refs: [
              { id: 'refid1', name: 'master', type: 'branch' },
              { id: 'refid2', name: 'v1.10.02', type: 'tag' }
            ]
          },
          worker: workersSample.results[0]
        }
        const version2 = {
          id: 'versionid2',
          configuration: {
            test: 'test'
          },
          revision: {
            id: 'revisionid2',
            state: 'processing',
            hash: 'yyyyyyyyyyyyyyyy',
            author: 'Patrick',
            message: 'merge message',
            created: '2020-05-27',
            commit_url: 'https://yyyyyyyyyyyyyyyy',
            refs: []
          },
          worker: workersSample.results[0]
        }
        const response = {
          number: 1,
          count: 2,
          next: null,
          results: [version1, version2]
        }

        mock.onGet('/workers/workerid/versions/', { page: 1 }).reply(200, response)
        await store.listVersions('workerid', { page: 1 })

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/workerid/versions/'
          }
        ])
        assert.deepStrictEqual(store.workerVersions, { versionid1: version1, versionid2: version2 })
      })
    })

    describe('getWorkerVersion', () => {
      it('gets a specific worker version', async () => {
        const reply = {
          id: 'versionid1',
          configuration: {
            test: 'test'
          },
          revision: {
            id: 'revisionid1',
            state: 'created',
            hash: 'xxxxxxxxxxxxxxxx',
            author: 'Bob',
            message: 'commit message',
            created: '2020-05-27',
            commit_url: 'https://xxxxxxxxxxxxxxxx',
            refs: [
              { id: 'refid1', name: 'master', type: 'branch' },
              { id: 'refid2', name: 'v1.10.02', type: 'tag' }
            ]
          },
          worker: workersSample.results[0]
        }

        mock.onGet('/workers/versions/versionid1/').reply(200, reply)

        await store.getWorkerVersion('versionid1')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/versions/versionid1/'
          }
        ])
        assert.deepStrictEqual(store.workerVersions, { versionid1: reply })
      })
    })

    describe('getWorker', () => {
      it('gets a specific worker', async () => {
        mock.onGet('/workers/workerid/').reply(200, workerSample)

        await store.getWorker('workerid')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/workers/workerid/'
          }
        ])
        assert.deepStrictEqual(store.workers, { workerid: workerSample })
      })
    })

    describe('getWorkerRun', () => {
      it('retrieves a specific worker run and stores the configuration', async () => {
        mock.onGet('/process/workers/workerRun2/').reply(200, workerRunsSample[1])

        await store.getWorkerRun('workerRun2')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/process/workers/workerRun2/'
          }
        ])

        assert.deepStrictEqual(store.workerRuns, {
          workerRun2: workerRunsSample[1]
        })
        assert.deepStrictEqual(store.workerConfigurations, {
          workerid2: {
            configid2: workerConfigurationsSample[1]
          }
        })
      })
    })

    describe('createLocalWorker', () => {
      it("creates a user's local worker", async () => {
        const payload = {
          name: workerSample.name,
          type: workerSample.type,
          slug: workerSample.slug
        }
        mock.onPost('/workers/', payload).reply(201, workerSample)

        await store.createWorker(payload)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/workers/'
          }
        ])
        assert.deepStrictEqual(store.workers, {
          [workerSample.id]: workerSample
        })
      })

      it('throws a specific error in case the worker already exists', async () => {
        const payload = {
          name: workerSample.name,
          type: workerSample.type,
          slug: workerSample.slug
        }
        mock.onPost('/workers/', payload).reply(200, workerSample)

        const err = await assertRejects(async () => { await store.createWorker(payload) })
        assert.ok(err instanceof DuplicatedWorker)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/workers/'
          }
        ])
        assert.deepStrictEqual(store.workers, {})
      })
    })

    describe('updateWorker', () => {
      it('updates attributes of a worker', async () => {
        store.workers = {
          [workerSample.id]: workerSample,
          b: {}
        }
        const payload = {
          name: 'A',
          type: 'new_type',
          slug: 'a',
          description: ''
        }
        mock.onPatch(`/workers/${workerSample.id}/`, payload).reply(200, { ...payload, id: workerSample.id })

        await store.updateWorker(workerSample.id, payload)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'patch',
            url: `/workers/${workerSample.id}/`
          }
        ])
        assert.deepStrictEqual(store.workers, {
          [workerSample.id]: { ...payload, id: workerSample.id },
          b: {}
        })
      })
    })

    describe('listUserWorkerRuns', () => {
      it("retrieves a User's local worker runs", async () => {
        const response = {
          results: workerRunsSample,
          count: 3
        }
        mock.onGet('/process/local/workers/', { page: 1 }).reply(200, response)

        await store.listUserWorkerRuns()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/process/local/workers/'
          }
        ])
        assert.deepStrictEqual(store.workerRuns, {
          workerRun1: workerRunsSample[0],
          workerRun2: workerRunsSample[1],
          workerRun3: workerRunsSample[2]
        })
      })
    })

    describe('createWorkerVersion', () => {
      it("creates a user's worker version", async () => {
        const payload = {
          configuration: {},
          model_usage: 'disabled',
          gpu_usage: 'disabled'
        }
        const sample = workerVersionsSample.results[0]
        mock.onPost('/workers/workerid/versions/', payload).reply(201, sample)

        await store.createWorkerVersion('workerid', payload)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/workers/workerid/versions/'
          }
        ])
        assert.deepStrictEqual(store.workerVersions, {
          [sample.id]: sample
        })
      })
    })

    describe('createUserWorkerRun', () => {
      it('creates a local worker run', async () => {
        store.workerRuns = {
          workerRun1: workerRunsSample[0]
        }
        const response = {
          id: 'local-worker-run-id',
          parents: [],
          worker_version: {
            id: 'some-version-id'
          },
          process: {
            name: 'My local process',
            id: 'user-local-process-id'
          },
          configuration: null,
          model_version: null,
          summary: 'Worker version @ sea'
        }
        mock.onPost('/process/local/workers/create/', { worker_version_id: 'some-version-id' }).reply(201, response)

        await store.createUserWorkerRun('some-version-id')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'post',
            url: '/process/local/workers/create/'
          }
        ])
        assert.deepStrictEqual(store.workerRuns, {
          workerRun1: workerRunsSample[0],
          'local-worker-run-id': response
        })
      })
    })
  })
})
