import { assert } from 'chai'
import { cloneDeep, set as lodashSet } from 'lodash'
import { createPinia, setActivePinia } from 'pinia'
import { createApp, markRaw } from 'vue'
import MockAdapter from 'axios-mock-adapter'
import { findHandler } from 'axios-mock-adapter/src/utils'

// A quick alternative for Node's assert.rejects, since assert 2.0.0 is not yet available in webpack.
export const assertRejects = async (fn, ...args) => {
  assert.strictEqual(typeof fn, 'function', 'assertRejects requires a function')
  try {
    await fn(...args)
  } catch (e) {
    return e
  }
  throw new Error('Promise was not rejected')
}

/**
 * A strange hack to add two missing features on the MockAdapter:
 *
 * - Add a `.history.all` attribute for all requests from any method, in order
 *   This allows both assertions on a method type, and assertion on every request made during the test, in order:
 *   assert.deepStrictEqual(mock.history.all, [{ method: 'post', data: '...' }, { method: 'get', params: { ... } }])
 * - Ensure that mock handlers are all called at least once, like what responses does in Python,
 *   by adding a `.calledHandlers` attribute, filling it when a request is performed, and checking it against
 *   `.handlers` upon calling `.reset()`.
 *
 * The adapter is pretty hard to extend as it does not use ES6 features to stay compatible with older browsers;
 * this implementation therefore uses a Proxy instance to override the constructor and 'install' all our changes
 * upon instantiation.
 *
 * To use this in a Mocha test, create a variable in your `describe` block, set it to `new FakeAxios(axios)`
 * in the `before` hook, then call `.reset()` in the `afterEach` hook and `.restore()` in the `after` hook.
 */
export const FakeAxios = new Proxy(MockAdapter, {
  construct (MockAdapter, args) {
    const mockAdapter = new MockAdapter(...args)

    /*
     * Override the MockAdapter.adapter function, which is what is called by Axios
     * and forwards requests to the MockAdapter, to 'note' that a handler has been called.
     *
     * Most of this is copied from axios-mock-adapter's own code as this is the lowest-level function
     * that can be overridden on the MockAdapter; everything else is not linked to `this`.
     * See https://github.com/ctimmerm/axios-mock-adapter/blob/master/src/handle_request.js
     */
    const originalAdapter = mockAdapter.adapter.bind(mockAdapter)()
    mockAdapter.adapter = () => config => {
      // Remove the base URL when set; allows mocking /api/v1 instead of http://instance/api/v1/…
      let url = config.url
      if (config.baseURL && url.startsWith(config.baseURL)) url = url.slice(config.baseURL.length)

      const handler = findHandler(
        mockAdapter.handlers,
        config.method,
        url,
        config.data,
        config.params,
        config.headers,
        config.baseURL
      )
      // Add the handler to called ones if not already done
      if (handler && !mockAdapter.calledHandlers[config.method].includes(handler)) {
        mockAdapter.calledHandlers[config.method].push(handler)
      }

      // No need to handle a case where there is no handler; the original adapter will deal with it
      return originalAdapter(config)
    }

    // Install the new adapter into the Axios instance
    mockAdapter.axiosInstance.defaults.adapter = mockAdapter.adapter()

    // Build the object from the keys of `.handlers` (the supported HTTP methods) with empty arrays
    mockAdapter.resetCalledHandlers = () => {
      mockAdapter.calledHandlers = Object.fromEntries(Object.keys(mockAdapter.handlers).map(key => ([key, []])))
    }

    /*
     * Now, override `.reset()` to check `.calledHandlers` against `.handlers` at the end of a test
     * and add support for `.history.all`. This has to be done on each reset as the history gets destroyed
     */
    const originalReset = mockAdapter.reset.bind(mockAdapter)
    mockAdapter.reset = () => {
      /*
       * This automatic check has some limitations when used with replyOnce as such handlers are unregistered once called.
       * In the below example the same request is completed twice.
       *
       * | Handlers              | Registered | Called   |
       * | --------------------- | ---------- | -------- |
       * | reply                 | 1          | 1        |
       * | replyOnce             | 0          | 1 (fail) |
       * | replyOnce → replyOnce | 0          | 2        |
       * | reply → replyOnce     | 2          | 1        |
       * | replyOnce → reply     | 1          | 2        |
       * | reply → reply         | 1          | 1        |
       */
      try {
        if (mockAdapter.handlers && mockAdapter.calledHandlers) {
          /*
           * Handlers are sorted in the order in which they were declared, but they might not always
           * be called in that exact same order; and most of the time, the ordering of API calls is meaningless.
           * This sorts all the arrays with Array.sort to ensure the assertion works even with unordered API calls.
           */
          const sortHandlers = handlers => Object.fromEntries(Object.entries(handlers).map(([key, value]) => [key, value.sort()]))
          assert.deepStrictEqual(
            sortHandlers(mockAdapter.calledHandlers),
            sortHandlers(mockAdapter.handlers),
            'Some response mocks were not called'
          )
        }
      } finally {
        /*
         * Reset anyway, even when this assertion fails,
         * to ensure other tests are not affected are developers are not confused
         */
        mockAdapter.resetCalledHandlers()
        originalReset()

        // Override Array.push to also push to history.all
        Object.values(mockAdapter.history).forEach(array => {
          const originalPush = array.push.bind(array)
          array.push = request => {
            const newRequest = { ...request }
            if (
              newRequest.headers &&
              newRequest.headers['Content-Type'] &&
              newRequest.headers['Content-Type'].includes('application/json')
            ) {
              // Parse the JSON data sent along with the request to make assertions over request bodies easier
              newRequest.data = JSON.parse(newRequest.data)
            }
            originalPush(newRequest)
            mockAdapter.history.all.push(newRequest)
          }
        })
        mockAdapter.history.all = []
      }
    }

    // Do a first reset to initialize everything properly
    mockAdapter.reset()

    return mockAdapter
  }
})

/**
 * Vuex store plugin to provide history, quick reset, and an equivalent of Python's `threading.Barrier` for actions.
 * To use it, just add `plugins: [StoreTestPlugin]` to a Vuex store config.
 * @param {Vuex.Store} store A Vuex Store instance.
 */
export const StoreTestPlugin = store => {
  // Provides a history of actions and mutations.
  store.history = []

  store.subscribe(event => {
    const historyItem = { mutation: event.type }
    if (event.payload !== undefined) historyItem.payload = cloneDeep(event.payload)
    store.history.push(historyItem)
  })

  store.subscribeAction({
    before (event) {
      const historyItem = { action: event.type }
      if (event.payload !== undefined) historyItem.payload = cloneDeep(event.payload)
      store.history.push(historyItem)
    }
  })

  // Keep a list of currently running action promises, to provide a `await store.actionsCompleted` and avoid async issues in tests
  store._promises = []
  const originalDispatch = store.dispatch.bind(store)
  store.dispatch = (...args) => {
    const promise = originalDispatch(...args)
    store._promises.push(promise)
    // Automatically remove once the promise ends even if it fails
    promise.finally(() => {
      const index = store._promises.indexOf(promise)
      if (index >= 0) store._promises.splice(index, 1)
    })
    return promise
  }

  store.actionsCompleted = () => Promise.all(store._promises)
    .catch(() => {
      // Ignore all errors; the point here is only to wait for all actions to end, no matter whether they are successful or not.
    })
    .then(() => {
      /*
       * When actionsCompleted is called, some promises might cause more actions to be called,
       * causing more promises to appear in thstoreis._promises; when this happens, recurse.
       */
      if (store._promises.length) return store.actionsCompleted()
    })

  // Provide a quick store reset to avoid leaky unit tests
  const initialState = cloneDeep(store.state)
  store.reset = () => {
    store.replaceState(cloneDeep(initialState))
    store._promises = []
    store.history = []
  }

  /**
   * Shortcut to edit the state outside of a mutation for easier test setup.
   * Avoids Vuex warnings about updating the state outside of a mutation.
   * Example: `setState('auth.user.email', 'user@domain.com')`
   *
   * @param {string} path Path to where the changes will be applied. Set to `''` to update the root state.
   * @param {any} value Any value to set at the given path using `_.set`.
   */
  store.setState = (path, value) => store._withCommit(() => lodashSet(store.state, path, value))
}

/**
 * A Pinia store plugin that adds a `.actionsCompleted()` helper to each store and to the Pinia instance,
 * allowing to wait for all Pinia actions to finish, successfully or not.
 *
 * Waiting for `store.actionsCompleted()` will only wait for the actions of this specific store.
 * `pinia.actionsCompleted()` will wait for all actions of all stores created with this Pinia instance.
 *
 * @param {import('pinia').PiniaPluginContext} param0
 */
export function actionsCompletedPlugin ({ pinia, store }) {
  store._promises = markRaw([])

  store.actionsCompleted = markRaw(() => Promise.allSettled(store._promises))

  store.$onAction(({ after, onError }) => {
    const promise = new Promise((resolve, reject) => {
      // ESLint thinks this after() call is a Mocha after hook and not the Pinia hook
      // eslint-disable-next-line mocha/no-top-level-hooks
      after(resolve)
      onError(reject)
    }).finally(() => {
      const index = store._promises.indexOf(promise)
      if (index >= 0) store._promises.splice(index, 1)
    })
    store._promises.push(promise)
  })

  if (!pinia.actionsCompleted) {
    pinia.actionsCompleted = markRaw(
      () => Promise.allSettled([...pinia._s.values()].map(store => store.actionsCompleted()))
    )
  }
}

/**
 * Set up a Vue app and a Pinia instance with the specified Pinia plugins, for use only in unit tests.
 * This should be called before calling any useStore functions.
 * Component tests should use `createTestingPinia` from the `@pinia/testing` package instead.
 *
 * @param {import('pinia').PiniaPlugin[]} plugins Array of plugins to use on the Pinia instance.
 * @returns {[import('pinia').Pinia, app: import('vue').App]} The created Pinia instance and Vue app.
 */
export function setUpTestPinia (plugins = []) {
  const app = createApp({})
  const pinia = createPinia()
  app.use(pinia)
  setActivePinia(pinia)
  plugins.forEach(plugin => pinia.use(plugin))
  return [pinia, app]
}
